#include "apue.h"
#include <fcntl.h>
#include <dirent.h>
#include <limits.h>

/* function type that is called for each filename */
typedef int Myfunc(const char *, const struct stat *, int);

static Myfunc   myfunc1;
static Myfunc   myfunc2;
static Myfunc   myfunc3;
static int      myftw(char *, Myfunc *);
static int      dopath(Myfunc *);
static int      cmpfile(const char *, const char *, const struct stat *);

static long nreg, ndir, nblk, nchr, nfifo, nslink, nsock, ntot;
static int type;
static long cnt4096;
static long cmpstrlen;
static char realpathname[PATH_MAX];
static char *cmpfilename, *cmpstr[MAXLINE];

int
main(int argc, char *argv[])
{
    int     ret;

    if (argc == 2) type = 1;
    else if (argc == 4 && strcmp(argv[2], "-comp") == 0) type = 2;
    else if (argc >= 4 && strcmp(argv[2], "-name") == 0) type = 3;
    else err_quit("usage:  myfind  <pathname>  [-comp <filename> | -name <str>...]");

    realpath(argv[1], realpathname);

    if (type == 1) ret = myftw(realpathname, myfunc1);      /* does it all */
    else if (type == 2) {
        struct stat statbuf;
        cmpfilename = argv[3];
        if (lstat(cmpfilename, &statbuf) < 0) err_sys("lstat error");
        cmpstrlen = statbuf.st_size;
        ret = myftw(realpathname, myfunc2);
    } else {
        for (int i = 3; i < argc; i++)
            cmpstr[i - 3] = argv[i];
        cmpstrlen = argc - 3;
        ret = myftw(realpathname, myfunc3);
    }

    if (type == 1) {
        ntot = nreg + ndir + nblk + nchr + nfifo + nslink + nsock;
        if (ntot == 0)
            ntot = 1;       /* avoid divide by 0; print 0 for all counts */
        printf("regular files                        = %7ld, %5.2f %%\n", nreg,
        nreg*100.0/ntot);
        printf("regular files not longer than 4096B  = %7ld, %5.2f %%\n", cnt4096,
        cnt4096*100.0/ntot);
        printf("directories                          = %7ld, %5.2f %%\n", ndir,
        ndir*100.0/ntot);
        printf("block special                        = %7ld, %5.2f %%\n", nblk,
        nblk*100.0/ntot);
        printf("char special                         = %7ld, %5.2f %%\n", nchr,
        nchr*100.0/ntot);
        printf("FIFOs                                = %7ld, %5.2f %%\n", nfifo,
        nfifo*100.0/ntot);
        printf("symbolic links                       = %7ld, %5.2f %%\n", nslink,
        nslink*100.0/ntot);
        printf("sockets                              = %7ld, %5.2f %%\n", nsock,
        nsock*100.0/ntot);
        exit(ret);
    }
}

/*
 * Descend through the hierarchy, starting at "pathname".
 * The caller's func() is called for every file.
 */
#define FTW_F   1       /* file other than directory */
#define FTW_D   2       /* directory */
#define FTW_DNR 3       /* directory that can't be read */
#define FTW_NS  4       /* file that we can't stat */

static char *fullpath;      /* contains full pathname for every file */
static size_t pathlen;

static int                  /* we return whatever func() returns */
myftw(char *pathname, Myfunc *func)
{
    fullpath = path_alloc(&pathlen);    /* malloc PATH_MAX+1 bytes */
                                        /* ({Prog pathalloc}) */
    if (pathlen <= strlen(pathname)) {
        pathlen = strlen(pathname) * 2;
        if ((fullpath = realloc(fullpath, pathlen)) == NULL)
            err_sys("realloc failed");
    }
    strcpy(fullpath, pathname);
    return(dopath(func));
}

/*
 * Descend through the hierarchy, starting at "fullpath".
 * If "fullpath" is anything other than a directory, we lstat() it,
 * call func(), and return.  For a directory, we call ourself
 * recursively for each name in the directory.
 */
static int                  /* we return whatever func() returns */
dopath(Myfunc* func)
{
    struct stat     statbuf;
    struct dirent   *dirp;
    DIR             *dp;
    int             ret, n;

    if (lstat(fullpath, &statbuf) < 0)  /* stat error */
        return(func(fullpath, &statbuf, FTW_NS));
    if (S_ISDIR(statbuf.st_mode) == 0)  /* not a directory */
        return(func(fullpath, &statbuf, FTW_F));

    /*
     * It's a directory.  First call func() for the directory,
     * then process each filename in the directory.
     */
    if ((ret = func(fullpath, &statbuf, FTW_D)) != 0)
        return(ret);

    n = strlen(fullpath);
    if (n + NAME_MAX + 2 > pathlen) {   /* expand path buffer */
        pathlen *= 2;
        if ((fullpath = realloc(fullpath, pathlen)) == NULL)
            err_sys("realloc failed");
    }
    fullpath[n++] = '/';
    fullpath[n] = 0;

    if ((dp = opendir(fullpath)) == NULL)   /* can't read directory */
        return(func(fullpath, &statbuf, FTW_DNR));

    while ((dirp = readdir(dp)) != NULL) {
        if (strcmp(dirp->d_name, ".") == 0  ||
            strcmp(dirp->d_name, "..") == 0)
                continue;       /* ignore dot and dot-dot */
        strcpy(&fullpath[n], dirp->d_name); /* append name after "/" */
        if ((ret = dopath(func)) != 0)      /* recursive */
            break;  /* time to leave */
    }
    fullpath[n-1] = 0;  /* erase everything from slash onward */

    if (closedir(dp) < 0)
        err_ret("can't close directory %s", fullpath);
    return(ret);
}

static int
myfunc1(const char *pathname, const struct stat *statptr, int type)
{
    switch (type) {
    case FTW_F:
        switch (statptr->st_mode & S_IFMT) {
        case S_IFREG:
            nreg++;
            if (statptr->st_size <= 4096)
                cnt4096++;
            break;
        case S_IFBLK:   nblk++;     break;
        case S_IFCHR:   nchr++;     break;
        case S_IFIFO:   nfifo++;    break;
        case S_IFLNK:   nslink++;   break;
        case S_IFSOCK:  nsock++;    break;
        case S_IFDIR:   /* directories should have type = FTW_D */
            err_dump("for S_IFDIR for %s", pathname);
        }
        break;
    case FTW_D:
        ndir++;
        break;
    case FTW_DNR:
        err_ret("can't read directory %s", pathname);
        break;
    case FTW_NS:
        err_ret("stat error for %s", pathname);
        break;
    default:
        err_dump("unknown type %d for pathname %s", type, pathname);
    }
    return(0);
}

static int myfunc2(const char *pathname, const struct stat *statptr, int type) {
    // fprintf(stderr, "myfunc2: %s\n", pathname);
    if (type != FTW_F || (statptr->st_mode & S_IFMT) != S_IFREG)
        return 0;
    if (cmpfile(pathname, cmpfilename, statptr) == 1)
        printf("%s\n", pathname);
    return 0;
}

static int myfunc3(const char *pathname, const struct stat *statptr, int type) {
    if (type != FTW_F || (statptr->st_mode & S_IFMT) != S_IFREG)
        return 0;
    const char *filename = pathname + strlen(pathname);
    while (filename >= pathname && *filename != '/')
        filename--;
    ++filename;
    for (int i = 0; i < cmpstrlen; ++i)
        if (strcmp(filename, cmpstr[i]) == 0) {
            printf("%s\n", pathname);
            return 0;
        }
    return 0;
}

static int cmpfile(const char *filename1, const char *filename2, const struct stat *statptr) {
    if (statptr->st_size != cmpstrlen)
        return 0;
    int fd1 = open(filename1, O_RDONLY);
    int fd2 = open(filename2, O_RDONLY);
    int n1, n2;
    static char buf1[BUFSIZ], buf2[BUFSIZ];
    if (fd1 < 0 || fd2 < 0)
        err_sys("open error");

    while ((n1 = read(fd1, buf1, BUFSIZ)) > 0 && (n2 = read(fd2, buf2, BUFSIZ)) > 0) {
        if (n1 != n2) goto diff;
        if (memcmp(buf1, buf2, n1) != 0) goto diff;
    }
    if (read(fd1, buf1, BUFSIZ) > 0 || read(fd2, buf2, BUFSIZ) > 0) goto diff;
    close(fd1), close(fd2);
    return 1;
diff:
    close(fd1), close(fd2);
    return 0;
}

